//
// Created by Tom on 07/03/2023.
//

#ifndef TYPON_BUILTINS_HPP
#define TYPON_BUILTINS_HPP

#include <fmt/format.h>
#include <iostream>
#include <optional>
#include <ostream>
#include <string>

#include <python/basedef.hpp>
#include <typon/typon.hpp>

#ifdef __cpp_lib_unreachable
#include <utility>
[[noreturn]] inline void TYPON_UNREACHABLE() { std::unreachable(); }
#else
#include <cstdlib>
[[noreturn]] inline void TYPON_UNREACHABLE() { std::abort(); }
#endif

#define COMMA() ,
#define METHOD(ret, name, args, ...)                                           \
  struct {                                                                     \
    ret operator() args __VA_ARGS__ type *self;                                \
  } name{this};

#define FUNCTION(ret, name, args, ...)                                         \
  struct {                                                                     \
    ret operator() args __VA_ARGS__                                            \
  } name;

using namespace std::literals;

template <typename T>
concept PyUserType = requires { typename T::type; };

template <typename T> struct RealType {
  using type = T;
};

template <PyUserType T> struct RealType<T> {
  using type = typename T::type;
};

template <typename T> using PyObj = std::shared_ptr<typename RealType<T>::type>;

template <typename T, typename... Args> auto pyobj(Args &&...args) -> PyObj<T> {
  return std::make_shared<typename RealType<T>::type>(
      std::forward<Args>(args)...);
}

// typon_len

template <typename T>
concept PyIterator = requires(T t) {
  { t.py_next() } -> std::same_as<std::optional<T>>;
};

template <typename T>
concept PyIterable = requires(T t) {
  { t.py_iter() } -> PyIterator;
};

template <PyIterable T, PyIterator U> U iter(const T &t) { return t.py_iter(); }

template <typename T>
concept CppSize = requires(const T &t) {
  { t.size() } -> std::same_as<size_t>;
};

template <typename T>
concept PyLen = requires(const T &t) {
  { t.py_len() } -> std::same_as<size_t>;
};

template <CppSize T>
  requires(!PyLen<T>)
size_t len(const T &t) {
  return t.size();
}

template <PyLen T> size_t len(const T &t) { return t.py_len(); }

template <typename T>
concept PyNext = requires(T t) {
  { t.py_next() } -> std::same_as<std::optional<typename T::value_type>>;
};

struct {
  template <PyNext T>
  std::optional<typename T::value_type>
  operator()(T &t, std::optional<typename T::value_type> def = std::nullopt) {
    auto opt = t.py_next();
    return opt ? opt : def;
  }
} next;

template <typename T>
std::ostream &operator<<(std::ostream &os, std::optional<T> const &opt) {
  return opt ? os << opt.value() : os << "None";
}

bool is_cpp() { return true; }

class NoneType {
public:
  template <typename T> operator T *() const { return nullptr; }

  template <typename T> operator std::optional<T>() const {
    return std::nullopt;
  }
} PyNone{};

#define system_error(err, message)                                             \
  do {                                                                         \
    puts(message);                                                             \
    throw fmt::system_error(err, message);                                     \
  } while (0)

#include "builtins/bool.hpp"
#include "builtins/complex.hpp"
#include "builtins/dict.hpp"
#include "builtins/list.hpp"
#include "builtins/print.hpp"
#include "builtins/range.hpp"
#include "builtins/set.hpp"
#include "builtins/str.hpp"

struct {
  struct type {
    METHOD(
        typon::Task<PyStr>, read, (size_t size = -1), const {
          if (size == -1) {
            size = self->len;
          }
          PyStr buf(size, '\0');
          int nbytes = co_await typon::io::read(self->fd, buf.data(), size);
          if (nbytes < 0) {
            system_error(-nbytes, "read()");
          }
          co_return std::move(buf);
        })
    METHOD(
        typon::Task<void>, close, (),
        const { co_await typon::io::close(self->fd); })

    type(int fd = -1, size_t len = 0) : fd(fd), len(len) {}

    type(const type &other)
        : fd(other.fd), len(other.len), read(this), close(this) {}

    int fd;
    size_t len;
  };
} file;

namespace typon {
using PyFile = decltype(file)::type;
}

typon::Task<typon::PyFile> open(const PyStr &path, std::string_view mode) {
  const char *path_c = path.c_str();
  struct statx statxbuf;
  if (int err = co_await typon::io::statx(AT_FDCWD, path_c, 0, STATX_SIZE, &statxbuf)) {
    system_error(-err, "statx()");
  }
  int fd = co_await typon::io::openat(AT_FDCWD, path_c, O_RDONLY, 0);
  if (fd < 0) {
    system_error(-fd, "openat()");
  }
  co_return typon::PyFile(fd, statxbuf.stx_size);
}

#include "../typon/generator.hpp"

#endif // TYPON_BUILTINS_HPP
